import abc
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from fandango.language.grammar.nodes.node import Node, NodeSettings


class HasSettings(abc.ABC):
    @abc.abstractmethod
    def settings_for(self, node: "Node") -> Optional["NodeSettings"]:
        raise NotImplementedError()



import pytest

from fandango import parse, FandangoParseError, DerivationTree
from fandango.constraints.comparison import ComparisonConstraint
from fandango.constraints.expression import ExpressionConstraint
from fandango.language.symbols import NonTerminal, Terminal


@pytest.mark.parametrize(
    "expression",
    [
        "f'test'",
        "f'{x}'",
        "f'{x} {y}'",
        "f'{x} {y} {z}'",
        "f'{x} {y} {z} {a}'",
        "f'{x} {y} {z} {a} {b}'",
        "f'{x} {y} {z} {a} {b} {c}'",
        "f'{x}{y}{z}{a}{b}{c}'",
        "f'{x!r}'",
        "f'{x!s}'",
        "f'{x!a}'",
        "f'{<x>!r}'",
        "f'{<x>:{y}}'",
        "f'{<x>!r:{y}}{z}'",
        'f""',
        'f"test {x}"',
        'f"test {x} {y}"',
        'f"""\ntest\n{x}\n"""',
        'f"""\ntest\n{x} {y}\n"""',
        "f'''\ntest\n{x}\n'''",
        "f'''\ntest\n{x} {y}\n'''",
        """f"{x}" 'test0' "test1" f'{y}' 'test2'""",
    ],
)
def test_fstrings(expression):
    try:
        grammar, constraints = parse(
            f'<start> ::= <x>\n<x> ::= "x"\nwhere {expression}'
        )
    except FandangoParseError:
        pytest.fail(f"Failed to parse expression: {expression}")
    else:
        assert (
            grammar is not None
        ), f"Grammar should not be None for expression: {expression}"
        assert (
            len(constraints) == 1
        ), f"Constraints should contain one item for expression: {expression}"
        assert isinstance(
            constraints[0], ExpressionConstraint
        ), f"Constraints should be an ExpressionConstraint for expression: {expression}"


def test_concreate_constraint():
    VALID = DerivationTree(
        NonTerminal("<start>"),
        [
            DerivationTree(NonTerminal("<x>"), [DerivationTree(Terminal("1"))]),
        ],
    )

    INVALID = DerivationTree(
        NonTerminal("<start>"),
        [
            DerivationTree(NonTerminal("<x>"), [DerivationTree(Terminal("2"))]),
        ],
    )

    raw_grammar = """
<start> ::= <x>
<x> ::= "1" | "2" | "3"
where f"{<x>!s:03}" == "100"
"""
    grammar, constraints = parse(raw_grammar)
    assert grammar is not None, "Grammar should not be None"
    assert len(constraints) == 1, "Constraints should contain one item"
    assert isinstance(
        constraints[0], ComparisonConstraint
    ), "Constraints should be a ComparisonConstraint"
    constraint: ComparisonConstraint = constraints[0]
    assert len(constraint.searches) == 1, "Constraint should have one search"
    tmp_var: str = ""
    for search in constraint.searches:
        tmp_var = search
    assert (
        eval(constraint.left, {tmp_var: "25"}) == "250"
    ), "Left side of comparison should evaluate to '250'"
    assert (
        eval(constraint.right) == "100"
    ), "Right side of comparison should evaluate to '100'"
    assert constraint.check(VALID), "Constraint should pass for VALID tree"
    assert not constraint.check(INVALID), "Constraint should fail for INVALID tree"



import random
import re
from types import NoneType
from typing import TYPE_CHECKING
from collections.abc import Sequence
import exrex
from fandango.errors import FandangoValueError
from fandango.language.grammar.has_settings import HasSettings
from fandango.language.grammar.nodes.node import Node, NodeType
from fandango.language.symbols import Terminal
from fandango.language.tree import DerivationTree
import fandango.language.grammar.nodes as nodes
from fandango.language.tree_value import TreeValueType
from fandango.logger import LOGGER

if TYPE_CHECKING:
    import fandango


class TerminalNode(Node):
    def __init__(
        self,
        symbol: Terminal,
        grammar_settings: Sequence[HasSettings],
    ):
        self.symbol = symbol
        super().__init__(
            NodeType.TERMINAL, grammar_settings, distance_to_completion=1.0
        )

    def fuzz(
        self,
        parent: DerivationTree,
        grammar: "fandango.language.grammar.grammar.Grammar",
        max_nodes: int = 100,
        in_message: bool = False,
    ):
        repetitions = 1
        # Gmutator mutation (1a)
        if random.random() < self.settings.get("terminal_should_repeat"):
            if random.random() < 0.5:
                repetitions = 0
                return  # nop, don't add a node
            else:
                repetitions = nodes.MAX_REPETITIONS
        for _ in range(repetitions):
            if not self.symbol.is_regex:
                parent.add_child(DerivationTree(self.symbol))
            else:

                def get_one(pattern: str) -> str:
                    # Gmutator mutation (3)
                    if random.random() < self.settings.get("invert_regex"):
                        attempts = 0
                        while attempts < self.settings.get("max_out_of_regex_tries"):
                            attempt = exrex.getone(".*")
                            if not re.match(pattern, attempt):
                                return attempt
                            attempts += 1
                        LOGGER.warning(
                            f"Failed to generate a non-matching regex: {pattern}, falling back to matching regex"
                        )

                    return exrex.getone(pattern)

                instance: str | bytes
                if self.symbol.is_type(TreeValueType.BYTES):
                    # Exrex can't do bytes, so we decode to str and back
                    pattern = self.symbol.value().to_string("latin-1")
                    instance = get_one(pattern).encode("latin-1")
                elif self.symbol.is_type(TreeValueType.STRING):
                    instance = get_one(str(self.symbol.value()))
                else:
                    raise FandangoValueError(
                        f"Unsupported type: {self.symbol.value().type_}"
                    )
                parent.add_child(DerivationTree(Terminal(instance)))

    def accept(
        self,
        visitor: "fandango.language.grammar.node_visitors.node_visitor.NodeVisitor",
    ):
        return visitor.visitTerminalNode(self)

    def format_as_spec(self) -> str:
        return self.symbol.format_as_spec()

    def __eq__(self, other):
        return isinstance(other, TerminalNode) and self.symbol == other.symbol

    def __hash__(self):
        return hash(self.symbol)



import abc
from typing import Generic, Optional, TypeVar
from fandango.language.grammar.nodes.alternative import Alternative
from fandango.language.grammar.nodes.char_set import CharSet
from fandango.language.grammar.nodes.concatenation import Concatenation
from fandango.language.grammar.nodes.node import Node
from fandango.language.grammar.nodes.non_terminal import NonTerminalNode
from fandango.language.grammar.nodes.repetition import Repetition, Star, Plus, Option
from fandango.language.grammar.nodes.terminal import TerminalNode


AggregateType = TypeVar("AggregateType")


class NodeVisitor(abc.ABC, Generic[AggregateType]):
    def visit(self, node: Node):
        return node.accept(self)

    def default_result(self) -> AggregateType:
        return None  # type: ignore[return-value]

    def aggregate_results(
        self, aggregate: AggregateType, result: Optional[Node]
    ) -> AggregateType:
        return aggregate

    def visitChildren(self, node: Node) -> AggregateType:
        # noinspection PyNoneFunctionAssignment
        result = self.default_result()
        for child in node.children():
            # noinspection PyNoneFunctionAssignment
            result = self.aggregate_results(result, self.visit(child))
        return result

    def visitAlternative(self, node: Alternative):
        return self.visitChildren(node)

    def visitConcatenation(self, node: Concatenation):
        return self.visitChildren(node)

    def visitRepetition(self, node: Repetition):
        return self.visit(node.node)

    def visitStar(self, node: Star):
        return self.visit(node.node)

    def visitPlus(self, node: Plus):
        return self.visit(node.node)

    def visitOption(self, node: Option):
        return self.visit(node.node)

    # noinspection PyUnusedLocal
    def visitNonTerminalNode(self, node: NonTerminalNode):
        return self.default_result()

    # noinspection PyUnusedLocal
    def visitTerminalNode(self, node: TerminalNode):
        return self.default_result()

    # noinspection PyUnusedLocal
    def visitCharSet(self, node: CharSet):
        return self.default_result()



from fandango.language.grammar.grammar import Grammar
from fandango.language.grammar.node_visitors.node_visitor import NodeVisitor
from fandango.language.grammar.nodes.non_terminal import NonTerminalNode
from fandango.language.symbols.non_terminal import NonTerminal


class MessageNestingDetector(NodeVisitor):
    def __init__(self, grammar: "Grammar"):
        self.grammar = grammar
        self.seen_nt: set[NonTerminal] = set()
        self.current_path: list[NonTerminal] = []

    def fail_on_nested_packet(self, start_symbol: NonTerminal):
        self.current_path.append(start_symbol)
        self.visit(self.grammar[start_symbol])
        self.current_path.pop()

    def visitNonTerminalNode(self, node: NonTerminalNode):
        if node.symbol not in self.seen_nt:
            self.seen_nt.add(node.symbol)
        elif node.sender is not None and node.symbol in self.current_path:
            str_path = [str(p) for p in self.current_path]
            raise RuntimeError(
                f"Found illegal packet-definitions within packet-definition of non_terminal {node.symbol}! DerivationPath: "
                + " -> ".join(str_path)
            )
        else:
            return

        if node.sender is not None:
            parties = self.grammar[node.symbol].msg_parties(include_recipients=False)
            if len(parties) != 0:
                raise RuntimeError(
                    f"Found illegal packet-definitions within packet-definition of non_terminal {node.symbol}: "
                    + ", ".join(parties)
                )
            return
        self.current_path.append(node.symbol)
        self.visit(self.grammar[node.symbol])
        self.current_path.pop()



from copy import deepcopy
from typing import Optional, Sequence

from fandango.errors import FandangoValueError
from fandango.language.grammar import ParsingMode
from fandango.language.grammar.parser.column import Column
from fandango.language.grammar.has_settings import HasSettings
from fandango.language.grammar.parser.iterative_parser import IterativeParser
from fandango.language.grammar.node_visitors.node_visitor import NodeVisitor
from fandango.language.grammar.nodes.char_set import CharSet
from fandango.language.grammar.parser.parse_state import ParseState
from fandango.language.grammar.grammar import Grammar
from fandango.language.grammar.nodes.node import Node
from fandango.language.grammar.nodes.non_terminal import NonTerminalNode
from fandango.language.grammar.nodes.terminal import TerminalNode
from fandango.language.grammar.nodes.concatenation import Concatenation
from fandango.language.grammar.nodes.alternative import Alternative
from fandango.language.grammar.nodes.repetition import Repetition, Option, Plus, Star
from fandango.language.symbols import Terminal, NonTerminal
from fandango.language.tree import DerivationTree
from fandango.language.tree_value import TreeValueType


class GrammarKeyError(KeyError):
    pass


class PathFinder(NodeVisitor):
    """
    For a given grammar and DerivationTree, this class
    finds possible upcoming message types, the nonterminals that generate them and the paths where the messages
    can be added to the DerivationTree.
    """

    def __init__(self, grammar: Grammar):
        self.grammar = grammar
        self.tree: Optional[DerivationTree] = None
        self.collapsed_tree: Optional[DerivationTree] = None
        self.current_tree: list[list[DerivationTree] | None] = []
        self.current_path: list[tuple[NonTerminal, bool]] = []
        self.result = PacketForecaster.ForecastingResult()

    def add_option(self, node: NonTerminalNode) -> None:
        mounting_path = PacketForecaster.MountingPath(
            self.collapsed_tree, tuple(self._collapsed_path(self.current_path))
        )
        f_packet = PacketForecaster.ForcastingPacket(node)
        f_packet.add_path(mounting_path)
        self.result.add_packet(node.sender, f_packet)

    @staticmethod
    def _collapsed_path(path: list[tuple[NonTerminal, bool]]):
        new_path = []
        for nt, new_node in path:
            if nt.is_type(TreeValueType.STRING) and str(nt.value()).startswith("<__"):
                continue
            elif nt.is_type(TreeValueType.BYTES) and bytes(nt.value()).startswith(
                b"<__"
            ):
                continue
            new_path.append((nt, new_node))
        return tuple(new_path)

    def find(
        self, tree: Optional[DerivationTree] = None
    ) -> "PacketForecaster.ForecastingResult":
        """
        Finds all possible protocol messages that can be mounted to the given DerivationTree.
        :param tree: The DerivationTree to base the search on. The DerivationTree must contain controlflow nodes
        as provided by the DerivationTree parser with the parsing option 'include_controlflow=True'
        """
        if tree is None:
            self.tree = DerivationTree(NonTerminal("<start>"))
        else:
            self.tree = tree
        self.collapsed_tree = self.grammar.collapse(self.tree)
        self.current_path = []
        self.current_tree = []

        self.result = PacketForecaster.ForecastingResult()
        self.current_path.append((self.tree.nonterminal, False))
        if len(self.tree.children) == 0:
            self.current_tree = [None]
        else:
            self.current_tree = [[self.tree.children[0]]]

        self.visit(self.grammar.rules[self.current_path[-1][0]])
        self.current_tree.pop()
        self.current_path.pop()
        return self.result

    def on_enter_controlflow(self, expected_nt: str):
        tree = self.current_tree[-1]
        cf_nt = (NonTerminal(expected_nt), True)
        if tree is not None:
            if len(tree) != 1:
                raise GrammarKeyError(
                    "Expected len(tree) == 1 for controlflow entries!"
                )
            assert isinstance(tree[0].symbol, NonTerminal)
            nt_name = tree[0].symbol.name()
            if nt_name != expected_nt:
                raise GrammarKeyError("Symbol mismatch!")
            cf_nt = (NonTerminal(nt_name), False)
        self.current_tree.append(None if tree is None else tree[0].children)
        self.current_path.append(cf_nt)

    def on_leave_controlflow(self):
        self.current_tree.pop()
        self.current_path.pop()

    def visitNonTerminalNode(self, node: NonTerminalNode):
        tree = self.current_tree[-1]
        if tree is not None:
            if tree[0].symbol != node.symbol:
                raise GrammarKeyError("Symbol mismatch")

        if node.sender is not None:
            if tree is None:
                self.add_option(node)
                return False
            else:
                return True
        self.current_tree.append(None if tree is None else tree[0].children)
        self.current_path.append((node.symbol, tree is None))
        try:
            result = self.visit(self.grammar.rules[node.symbol])
        finally:
            self.current_path.pop()
            self.current_tree.pop()
        return result

    def visitTerminalNode(self, node: TerminalNode):
        raise FandangoValueError(
            "PacketForecaster reached TerminalNode! This is a bug."
        )

    def visitConcatenation(self, node: Concatenation):
        self.on_enter_controlflow(f"<__{node.id}>")
        tree = self.current_tree[-1]
        child_idx = 0 if tree is None else (len(tree) - 1)
        continue_exploring = True
        if tree is not None:
            self.current_tree.append([tree[child_idx]])
            try:
                if len(node.nodes) <= child_idx:
                    raise GrammarKeyError(
                        "Tree contains more children, then concatination node"
                    )
                continue_exploring = self.visit(node.nodes[child_idx])
                child_idx += 1
            finally:
                self.current_tree.pop()
        while continue_exploring and child_idx < len(node.children()):
            next_child = node.children()[child_idx]
            self.current_tree.append(None)
            continue_exploring = self.visit(next_child)
            self.current_tree.pop()
            child_idx += 1
        self.on_leave_controlflow()
        return continue_exploring

    def visitAlternative(self, node: Alternative):
        self.on_enter_controlflow(f"<__{node.id}>")
        tree = self.current_tree[-1]

        if tree is not None:
            continue_exploring = True
            self.current_tree.append([tree[0]])
            fallback_tree = list(self.current_tree)
            fallback_path = list(self.current_path)
            found = False
            for alt in node.alternatives:
                try:
                    continue_exploring = self.visit(alt)
                    found = True
                    break
                except GrammarKeyError:
                    self.current_tree = fallback_tree
                    self.current_path = fallback_path
            self.current_tree.pop()
            self.on_leave_controlflow()
            if not found:
                raise GrammarKeyError("Alternative mismatch")
            return continue_exploring
        else:
            continue_exploring = False
            self.current_tree.append(None)
            for alt in node.alternatives:
                continue_exploring |= self.visit(alt)
            self.current_tree.pop()
            self.on_leave_controlflow()
            return continue_exploring

    def visitRepetition(self, node: Repetition):
        self.on_enter_controlflow(f"<__{node.id}>")
        ret = self.visitRepetitionType(node)
        self.on_leave_controlflow()
        return ret

    def visitRepetitionType(self, node: Repetition):
        tree = self.current_tree[-1]
        continue_exploring = True
        tree_len = 0
        if tree is not None and len(tree) != 0:
            tree_len = len(tree)
            self.current_tree.append([tree[-1]])
            continue_exploring = self.visit(node.node)
            self.current_tree.pop()

        # TODO match new computed length repetitions
        rep_min = node.min
        rep_max = node.max
        if node.bounds_constraint:
            prefix_tree = None
            for tree_list in self.current_tree[::-1]:
                if tree_list is None or len(tree_list) != 0:
                    continue
                prefix_tree = tree_list[-1].prefix()
                prefix_tree = self.grammar.collapse(prefix_tree.get_root())
                break
            assert prefix_tree is not None
            rep_min, _ = node.bounds_constraint.min(prefix_tree)
            rep_max, _ = node.bounds_constraint.max(prefix_tree)
        if continue_exploring and tree_len < rep_max:
            self.current_tree.append(None)
            continue_exploring = self.visit(node.node)
            self.current_tree.pop()
            if continue_exploring:
                return continue_exploring
        # TODO match new computed length repetitions
        if tree_len >= rep_min:
            return True
        return continue_exploring

    def visitStar(self, node: Star):
        self.on_enter_controlflow(f"<__{node.id}>")
        ret = self.visitRepetitionType(node)
        self.on_leave_controlflow()
        return ret

    def visitPlus(self, node: Plus):
        self.on_enter_controlflow(f"<__{node.id}>")
        ret = self.visitRepetitionType(node)
        self.on_leave_controlflow()
        return ret

    def visitOption(self, node: Option):
        self.on_enter_controlflow(f"<__{node.id}>")
        ret = self.visitRepetitionType(node)
        self.on_leave_controlflow()
        return ret


class PacketForecaster:
    class MountingPath:
        def __init__(
            self,
            tree: Optional[DerivationTree],
            path: tuple[tuple[NonTerminal, bool], ...],
        ):
            """
            Represents a path in the given DerivationTree where a protocol message can be mounted.
            """
            self.tree = tree
            self.path = path

        def __hash__(self):
            return hash((hash(self.tree), hash(self.path)))

        def __eq__(self, other):
            return hash(self) == hash(other)

        def __repr__(self):
            return f"({', '.join([f'({nt.format_as_spec()}, {new_node})' for nt, new_node in self.path])})"

    class ForcastingPacket:
        def __init__(self, node: NonTerminalNode):
            self.node = node
            self.paths: set[PacketForecaster.MountingPath] = set()

        def add_path(self, path: "PacketForecaster.MountingPath") -> None:
            self.paths.add(path)

    class ForcastingNonTerminals:
        def __init__(self):
            self.nt_to_packet = dict[NonTerminal, PacketForecaster.ForcastingPacket]()

        def get_non_terminals(self) -> set[NonTerminal]:
            return set(self.nt_to_packet.keys())

        def __getitem__(self, item: NonTerminal):
            return self.nt_to_packet[item]

        def add_packet(self, packet: "PacketForecaster.ForcastingPacket") -> None:
            """
            Adds a packet to the ForcastingNonTerminals.
            """
            if packet.node.symbol in self.nt_to_packet.keys():
                for path in packet.paths:
                    self.nt_to_packet[packet.node.symbol].add_path(path)
            else:
                self.nt_to_packet[packet.node.symbol] = packet

    class ForecastingResult:
        def __init__(self):
            self.parties_to_packets = dict[
                str, PacketForecaster.ForcastingNonTerminals
            ]()

        def get_msg_parties(self) -> set[str]:
            return set(self.parties_to_packets.keys())

        def contains_any_party(self, parties: list[str]):
            """
            Checks if the ForecastingResult contains any of the specified parties.
            :param parties: List of party names to check.
            :return: True if any party is found, False otherwise.
            """
            return any(party in self.parties_to_packets for party in parties)

        def __getitem__(self, item: str):
            return self.parties_to_packets[item]

        def __contains__(self, item: str) -> bool:
            return item in self.parties_to_packets

        def add_packet(
            self, party: Optional[str], packet: "PacketForecaster.ForcastingPacket"
        ) -> None:
            """
            Adds a packet to the ForecastingResult under the specified party.
            """
            if party is None:
                raise FandangoValueError("Party cannot be None")
            if party not in self.parties_to_packets.keys():
                self.parties_to_packets[party] = (
                    PacketForecaster.ForcastingNonTerminals()
                )
            self.parties_to_packets[party].add_packet(packet)

        def union(
            self, other: "PacketForecaster.ForecastingResult"
        ) -> "PacketForecaster.ForecastingResult":
            """
            Combines two ForecastingResults by adding all packets from the other result.
            Returns a copy of the current ForecastingResult with the combined packets.
            :param other: The other ForecastingResult to combine with.
            """
            c_new = deepcopy(self)
            c_other = deepcopy(other)
            for party, fnt in c_other.parties_to_packets.items():
                for fp in fnt.nt_to_packet.values():
                    c_new.add_packet(party, fp)
            return c_new

    class GrammarReducer(NodeVisitor):
        """
        Converts a grammar into a reduced form, where all protocol message defining NonTerminalNodes are replaced with
        a TerminalNode that describes the protocol message type.
        Message defining NonTerminals are replaced with a Terminal, in the form of <_packet_<message_type>>.
        This allows the PacketForecaster to predict upcoming
        protocol messages without parsing each protocol message again.
        """

        def __init__(self, grammar_settings: Sequence[HasSettings]):
            self._grammar_settings = grammar_settings
            self._reduced: dict[NonTerminal, Node] = dict()
            self.seen_keys: set[NonTerminal] = set()
            self.processed_keys: set[NonTerminal] = set()

        def process(self, grammar: Grammar) -> dict[NonTerminal, Node]:
            """
            Applies the grammar reduction to the provided grammar.
            """
            self._reduced = dict()
            self.seen_keys = set()
            self.seen_keys.add(NonTerminal("<start>"))
            self.processed_keys = set()
            diff_keys = self.seen_keys - self.processed_keys
            while len(diff_keys) != 0:
                key = diff_keys.pop()
                self._reduced[key] = self.visit(grammar.rules[key])
                self.processed_keys.add(key)
                diff_keys = self.seen_keys - self.processed_keys
            return self._reduced

        def default_result(self):
            return []

        def aggregate_results(self, aggregate, result):
            aggregate.append(result)
            return aggregate

        def visitConcatenation(self, node: Concatenation):
            return Concatenation(
                self.visitChildren(node),
                self._grammar_settings,
                node.id,
            )

        def visitTerminalNode(self, node: TerminalNode):
            return TerminalNode(node.symbol, self._grammar_settings)

        def visitAlternative(self, node: Alternative):
            return Alternative(
                self.visitChildren(node),
                self._grammar_settings,
                node.id,
            )

        def visitRepetition(self, node: Repetition):
            return Repetition(
                self.visit(node.node),
                self._grammar_settings,
                node.id,
                node.min,
                node.internal_max,
            )

        def visitOption(self, node: Option):
            return Option(
                self.visit(node.node),
                self._grammar_settings,
                node.id,
            )

        def visitPlus(self, node: Plus):
            return Plus(self.visit(node.node), self._grammar_settings, node.id)

        def visitStar(self, node: Star):
            return Star(
                self.visit(node.node),
                self._grammar_settings,
                node.id,
            )

        def visitCharSet(self, node: CharSet):
            return CharSet(node.chars, self._grammar_settings)

        def visitNonTerminalNode(self, node: NonTerminalNode):
            if node.sender is None and node.recipient is None:
                self.seen_keys.add(node.symbol)
                return node

            if node.symbol.is_type(TreeValueType.STRING):
                symbol = NonTerminal("<_packet_" + node.symbol.name()[1:])
            else:
                raise FandangoValueError("NonTerminal symbol must be a string!")
            repl_node = NonTerminalNode(
                symbol,
                self._grammar_settings,
                node.sender,
                node.recipient,
            )
            self._reduced[symbol] = TerminalNode(
                Terminal(node.symbol.value()), self._grammar_settings
            )
            self.seen_keys.add(symbol)
            self.processed_keys.add(symbol)
            return repl_node

    class PacketIterativeParser(IterativeParser):
        def __init__(self, grammar_rules: dict[NonTerminal, Node]):
            super().__init__(grammar_rules)
            self.reference_tree: Optional[DerivationTree] = None
            self.detailed_tree: Optional[DerivationTree] = None

        def construct_incomplete_tree(
            self, state: ParseState, table: list[Column]
        ) -> DerivationTree:
            i_tree = super().construct_incomplete_tree(state, table)
            i_cpy = deepcopy(i_tree)
            if self.reference_tree is None:
                raise FandangoValueError(
                    "Reference tree must be set before constructing the incomplete tree!"
                )
            for i_msg, r_msg in zip(
                i_cpy.protocol_msgs(), self.reference_tree.protocol_msgs()
            ):
                i_msg.msg.set_children(r_msg.msg.children)
                i_msg.msg.sources = r_msg.msg.sources
                symbol = r_msg.msg.symbol
                if isinstance(symbol, NonTerminal):
                    # TODO: Is this just to create a new string?
                    i_msg.msg.symbol = NonTerminal("<" + symbol.name()[1:])
                else:
                    raise FandangoValueError("NonTerminal symbol must be a string!")
            return i_cpy

    def __init__(self, grammar: Grammar):
        g_globals, g_locals = grammar.get_spec_env()
        reduced = PacketForecaster.GrammarReducer(grammar.grammar_settings).process(
            grammar
        )
        self.grammar = grammar
        self.reduced_grammar = Grammar(
            grammar.grammar_settings,
            reduced,
            grammar.fuzzing_mode,
            g_locals,
            g_globals,
        )
        self._parser = PacketForecaster.PacketIterativeParser(
            self.reduced_grammar.rules
        )

    def predict(self, tree: DerivationTree) -> "ForecastingResult":
        """
        Predicts the next possible message types based on the provided tree and the grammar,
        that the PacketForecaster was initialized with.
        :param tree: The DerivationTree to base the prediction on.
        """
        history_nts = ""
        for r_msg in tree.protocol_msgs():
            assert isinstance(r_msg.msg.symbol, NonTerminal)
            history_nts += r_msg.msg.symbol.name()
        self._parser.detailed_tree = tree

        finder = PathFinder(self.grammar)
        options = PacketForecaster.ForecastingResult()
        if history_nts == "":
            options = options.union(finder.find())
        else:
            self._parser.reference_tree = tree
            self._parser.new_parse(NonTerminal("<start>"), ParsingMode.INCOMPLETE)
            for suggested_tree in self._parser.consume(history_nts):
                for orig_r_msg, r_msg in zip(
                    tree.protocol_msgs(), suggested_tree.protocol_msgs()
                ):
                    assert isinstance(r_msg.msg.symbol, NonTerminal)
                    assert isinstance(orig_r_msg.msg.symbol, NonTerminal)
                    if (
                        r_msg.msg.symbol.name()[9:] == orig_r_msg.msg.symbol.name()[1:]
                        and r_msg.sender == orig_r_msg.sender
                        and r_msg.recipient == orig_r_msg.recipient
                    ):
                        cpy = orig_r_msg.msg.deepcopy(copy_parent=False)
                        assert isinstance(cpy.symbol, NonTerminal)
                        r_msg.msg.set_children(cpy.children)
                        r_msg.msg.sources = deepcopy(cpy.sources)
                        r_msg.msg.symbol = NonTerminal("<" + cpy.symbol.name()[1:])
                    else:
                        break
                else:
                    options = options.union(finder.find(suggested_tree))
        return options



import re


class LiteralGenerator:
    def __init__(self, call: str, nonterminals: dict):
        self.call = call
        self.nonterminals = nonterminals

    def __repr__(self):
        return f"LiteralGenerator({self.call!r}, {self.nonterminals!r})"

    def __str__(self):
        # Generators are created with internal variables;
        # we replace them with "..." to avoid cluttering the output.
        s = re.sub(r"___[0-9a-zA-Z_]+___", r"...", str(self.call))
        return s

    def __eq__(self, other):
        return (
            isinstance(other, LiteralGenerator)
            and self.call == other.call
            and self.nonterminals == other.nonterminals
        )

    def __hash__(self):
        return hash(self.call) ^ hash(self.nonterminals)



#!/usr/bin/env pytest

import unittest
from typing import Optional

from fandango.language.grammar import ParsingMode
from fandango.language.grammar.grammar import Grammar
from fandango.language.grammar.nodes.alternative import Alternative
from fandango.language.grammar.parser.iterative_parser import IterativeParser
from fandango.language.grammar.parser.parser import Parser
from fandango.language.parse import parse
from fandango.language.symbols import NonTerminal, Terminal
from fandango.language.tree import DerivationTree
from .utils import RESOURCES_ROOT, DOCS_ROOT, run_command


class IterParsingTester(Parser):
    def _parse_forest(
        self,
        word: str | bytes,
        start: str | NonTerminal = "<start>",
        *,
        mode: ParsingMode = ParsingMode.COMPLETE,
        hookin_parent: Optional[DerivationTree] = None,
        starter_bit=-1,
    ):
        self._iter_parser.new_parse(start, mode, hookin_parent, starter_bit)
        for char in word[:-1]:
            next(self._iter_parser.consume(char), None)
        yield from self._iter_parser.consume(word[-1])


class ParserTests(unittest.TestCase):
    # Type annotation for instance attribute set in setUp
    grammar: Grammar

    def setUp(self):
        with open(RESOURCES_ROOT / "fandango.fan") as file:
            grammar, _ = parse(file, use_stdlib=False, use_cache=False)
            assert grammar is not None
            self.grammar = grammar

    def test_rules(self):
        self.assertEqual(len(self.grammar._parser._iter_parser._rules), 9)
        self.assertEqual(len(self.grammar._parser._iter_parser._implicit_rules), 1)
        self.assertEqual(
            {((NonTerminal("<number>"), frozenset()),)},
            self.grammar._parser._iter_parser._rules[NonTerminal("<start>")],
        )
        alt_1 = self.grammar.rules[NonTerminal("<number>")]
        assert isinstance(alt_1, Alternative)
        alt_2 = self.grammar.rules[NonTerminal("<non_zero>")]
        assert isinstance(alt_2, Alternative)
        alt_3 = self.grammar.rules[NonTerminal("<digit>")]
        assert isinstance(alt_3, Alternative)
        concat_1 = alt_1.children()[0]
        star_1 = concat_1.children()[1]

        self.assertEqual(
            {((NonTerminal(f"<__{alt_1.id}>"), frozenset()),)},
            self.grammar._parser._iter_parser._rules[NonTerminal("<number>")],
        )
        self.assertEqual(
            {((NonTerminal(f"<__{alt_2.id}>"), frozenset()),)},
            self.grammar._parser._iter_parser._rules[NonTerminal("<non_zero>")],
        )
        self.assertEqual(
            {((NonTerminal(f"<__{alt_3.id}>"), frozenset()),)},
            self.grammar._parser._iter_parser._rules[NonTerminal("<digit>")],
        )
        self.assertEqual(
            {((NonTerminal("<*0*>"), frozenset()),)},
            self.grammar._parser._iter_parser._rules[NonTerminal(f"<__{star_1.id}>")],
        )
        self.assertEqual(
            {
                (
                    (NonTerminal("<non_zero>"), frozenset()),
                    (NonTerminal(f"<__{star_1.id}>"), frozenset()),
                )
            },
            self.grammar._parser._iter_parser._rules[NonTerminal(f"<__{concat_1.id}>")],
        )
        self.assertEqual(
            {
                ((Terminal("0"), frozenset()),),
                (
                    (
                        NonTerminal(f"<__{concat_1.id}>"),
                        frozenset(),
                    ),
                ),
            },
            self.grammar._parser._iter_parser._rules[NonTerminal(f"<__{alt_1.id}>")],
        )
        self.assertEqual(
            {
                ((Terminal("1"), frozenset()),),
                ((Terminal("2"), frozenset()),),
                ((Terminal("3"), frozenset()),),
                ((Terminal("4"), frozenset()),),
                ((Terminal("5"), frozenset()),),
                ((Terminal("6"), frozenset()),),
                ((Terminal("7"), frozenset()),),
                ((Terminal("8"), frozenset()),),
                ((Terminal("9"), frozenset()),),
            },
            self.grammar._parser._iter_parser._rules[NonTerminal(f"<__{alt_2.id}>")],
        )
        self.assertEqual(
            {
                ((Terminal("0"), frozenset()),),
                ((Terminal("1"), frozenset()),),
                ((Terminal("2"), frozenset()),),
                ((Terminal("3"), frozenset()),),
                ((Terminal("4"), frozenset()),),
                ((Terminal("5"), frozenset()),),
                ((Terminal("6"), frozenset()),),
                ((Terminal("7"), frozenset()),),
                ((Terminal("8"), frozenset()),),
                ((Terminal("9"), frozenset()),),
            },
            self.grammar._parser._iter_parser._rules[NonTerminal(f"<__{alt_3.id}>")],
        )


class TestComplexParsing(unittest.TestCase):
    # Type annotation for instance attribute set in setUp
    grammar: Grammar

    def setUp(self):
        with open(RESOURCES_ROOT / "constraints.fan") as file:
            grammar, _ = parse(file, use_stdlib=False, use_cache=False)
            assert grammar is not None
            self.grammar = grammar
            self.parser = Parser(grammar.rules)
            self.iter_parser = IterParsingTester(grammar.rules)

    def _test(self, example, tree):
        for parser in [self.parser, self.iter_parser]:
            actual_tree = parser.parse(example, "<ab>")
            self.assertEqual(tree, actual_tree)

    def test_bb(self):
        self._test(
            "bb",
            DerivationTree(
                NonTerminal("<ab>"),
                [
                    DerivationTree(
                        NonTerminal("<ab>"),
                        [
                            DerivationTree(
                                NonTerminal("<ab>"), [DerivationTree(Terminal(""))]
                            ),
                            DerivationTree(Terminal("b")),
                        ],
                    ),
                    DerivationTree(Terminal("b")),
                ],
            ),
        )

    def test_b(self):
        self._test(
            "b",
            DerivationTree(
                NonTerminal("<ab>"),
                [
                    DerivationTree(NonTerminal("<ab>"), [DerivationTree(Terminal(""))]),
                    DerivationTree(Terminal("b")),
                ],
            ),
        )

    def test_ab(self):
        self._test(
            "ab",
            DerivationTree(
                NonTerminal("<ab>"),
                [
                    DerivationTree(
                        NonTerminal("<ab>"),
                        [
                            DerivationTree(Terminal("a")),
                            DerivationTree(
                                NonTerminal("<ab>"), [DerivationTree(Terminal(""))]
                            ),
                        ],
                    ),
                    DerivationTree(Terminal("b")),
                ],
            ),
        )

    def test_a(self):
        self._test(
            "a",
            DerivationTree(
                NonTerminal("<ab>"),
                [
                    DerivationTree(Terminal("a")),
                    DerivationTree(NonTerminal("<ab>"), [DerivationTree(Terminal(""))]),
                ],
            ),
        )


class TestIncompleteParsing(unittest.TestCase):
    # Type annotation for instance attribute set in setUp
    grammar: Grammar

    def setUp(self):
        with open(RESOURCES_ROOT / "incomplete.fan") as file:
            grammar, _ = parse(file, use_stdlib=False, use_cache=False)
            assert grammar is not None
            self.grammar = grammar
            self.parser = Parser(grammar.rules)
            self.iter_parser = IterParsingTester(grammar.rules)

    def _test(self, example, tree):
        for parser in [self.parser, self.iter_parser]:
            parsed = False
            for actual_tree in parser.parse_multiple(
                example, "<start>", mode=ParsingMode.INCOMPLETE
            ):
                self.assertEqual(tree, actual_tree)
                parsed = True
                break
            self.assertTrue(parsed)

    def test_a(self):
        self._test(
            "aa",
            DerivationTree(
                NonTerminal("<start>"),
                [
                    DerivationTree(
                        NonTerminal("<ab>"),
                        [
                            DerivationTree(Terminal("a")),
                            DerivationTree(
                                NonTerminal("<ab>"), [DerivationTree(Terminal("a"))]
                            ),
                        ],
                    )
                ],
            ),
        )

    def test_regex(self):
        self._test(
            "ii",
            DerivationTree(
                NonTerminal("<start>"),
                [
                    DerivationTree(
                        NonTerminal("<c>"),
                        [
                            DerivationTree(Terminal("ii")),
                        ],
                    ),
                ],
            ),
        )


class TestDynamicRepetitionParsing(unittest.TestCase):
    # Type annotation for instance attribute set in setUp
    grammar: Grammar

    def setUp(self):
        with open(RESOURCES_ROOT / "dynamic_repetition.fan") as file:
            grammar, _ = parse(file, use_stdlib=False, use_cache=False)
            assert grammar is not None
            self.grammar = grammar
            self.parser = Parser(grammar.rules)
            self.iter_parser = IterParsingTester(grammar.rules)

    def _test(self, example, tree):
        for parser in [self.parser, self.iter_parser]:
            parsed = False
            for actual_tree in parser.parse_multiple(
                example, mode=ParsingMode.COMPLETE
            ):
                self.assertEqual(tree, actual_tree)
                parsed = True
                break
            self.assertTrue(parsed)

    def test_nested(self):
        self._test(
            "2(3aaa2bb)",
            DerivationTree(
                NonTerminal("<start>"),
                [
                    DerivationTree(
                        NonTerminal("<len>"),
                        [
                            DerivationTree(
                                NonTerminal("<number>"),
                                [
                                    DerivationTree(
                                        NonTerminal("<number_start>"),
                                        [DerivationTree(Terminal("2"))],
                                    )
                                ],
                            )
                        ],
                    ),
                    DerivationTree(Terminal("(")),
                    DerivationTree(
                        NonTerminal("<inner>"),
                        [
                            DerivationTree(
                                NonTerminal("<len>"),
                                [
                                    DerivationTree(
                                        NonTerminal("<number>"),
                                        [
                                            DerivationTree(
                                                NonTerminal("<number_start>"),
                                                [DerivationTree(Terminal("3"))],
                                            )
                                        ],
                                    )
                                ],
                            ),
                            DerivationTree(
                                NonTerminal("<letter>"), [DerivationTree(Terminal("a"))]
                            ),
                            DerivationTree(
                                NonTerminal("<letter>"), [DerivationTree(Terminal("a"))]
                            ),
                            DerivationTree(
                                NonTerminal("<letter>"), [DerivationTree(Terminal("a"))]
                            ),
                        ],
                    ),
                    DerivationTree(
                        NonTerminal("<inner>"),
                        [
                            DerivationTree(
                                NonTerminal("<len>"),
                                [
                                    DerivationTree(
                                        NonTerminal("<number>"),
                                        [
                                            DerivationTree(
                                                NonTerminal("<number_start>"),
                                                [DerivationTree(Terminal("2"))],
                                            )
                                        ],
                                    )
                                ],
                            ),
                            DerivationTree(
                                NonTerminal("<letter>"), [DerivationTree(Terminal("b"))]
                            ),
                            DerivationTree(
                                NonTerminal("<letter>"), [DerivationTree(Terminal("b"))]
                            ),
                        ],
                    ),
                    DerivationTree(Terminal(")")),
                ],
            ),
        )


class TestEmptyParsing(unittest.TestCase):
    # Type annotation for instance attribute set in setUp
    grammar: Grammar

    def setUp(self):
        with open(RESOURCES_ROOT / "empty.fan") as file:
            grammar, _ = parse(file, use_stdlib=False, use_cache=False)
            assert grammar is not None
            self.grammar = grammar
            self.parser = Parser(grammar.rules)
            self.iter_parser = IterParsingTester(grammar.rules)

    def _test(self, example: str, tree: DerivationTree):
        parsers: list[Parser] = [self.parser, self.iter_parser]
        for parser in parsers:
            actual_tree = parser.parse(example)
            print(type(parser), type(actual_tree))
            self.assertEqual(tree, actual_tree)

    def test_a(self):
        self._test(
            "1234",
            DerivationTree(
                NonTerminal("<start>"),
                [
                    DerivationTree(Terminal("123")),
                    DerivationTree(
                        NonTerminal("<digit>"), [DerivationTree(Terminal("4"))]
                    ),
                ],
            ),
        )

    def test_b(self):
        self._test(
            "123456",
            DerivationTree(
                NonTerminal("<start>"),
                [
                    DerivationTree(Terminal("12345")),
                    DerivationTree(Terminal("")),
                    DerivationTree(
                        NonTerminal("<digit>"), [DerivationTree(Terminal("6"))]
                    ),
                ],
            ),
        )


class TestCanContinueParsing(unittest.TestCase):

    def setUp(self):
        with open(RESOURCES_ROOT / "rgb.fan") as file:
            grammar, _ = parse(file, use_stdlib=False, use_cache=False)
            assert grammar is not None
            self.grammar = grammar
            self.iter_parser = IterativeParser(self.grammar.rules)

    def test_1(self):
        self.iter_parser.new_parse()
        next(self.iter_parser.consume(b"r"), None)
        self.assertTrue(self.iter_parser.can_continue())
        next(self.iter_parser.consume(b"g"), None)
        self.assertTrue(self.iter_parser.can_continue())
        next(self.iter_parser.consume(b"b"), None)
        self.assertTrue(self.iter_parser.can_continue())
        next(self.iter_parser.consume(b"d"), None)
        self.assertTrue(self.iter_parser.can_continue())
        next(self.iter_parser.consume(b";"), None)
        self.assertFalse(self.iter_parser.can_continue())

        self.iter_parser.new_parse()
        next(self.iter_parser.consume(b"rgbd;"), None)


class TestCLIParsing(unittest.TestCase):
    pass


class TestRegexParsing(TestCLIParsing):
    def test_infinity_abc(self):
        command = [
            "fandango",
            "parse",
            "-f",
            str(DOCS_ROOT / "infinity.fan"),
            "--validate",
            str(RESOURCES_ROOT / "abc.txt"),
            "--validate",
        ]
        out, err, code = run_command(command)
        self.assertEqual("", err)
        self.assertEqual("", out)
        self.assertEqual(0, code)

    def test_infinity_abcabc(self):
        command = [
            "fandango",
            "parse",
            "-f",
            str(DOCS_ROOT / "infinity.fan"),
            "--validate",
            str(RESOURCES_ROOT / "abcabc.txt"),
            "--validate",
        ]
        out, err, code = run_command(command)
        self.assertEqual("", err)
        self.assertEqual("", out)
        self.assertEqual(0, code)

    def test_infinity_abcd(self):
        # This should be rejected by the grammar
        command = [
            "fandango",
            "parse",
            "-f",
            str(DOCS_ROOT / "infinity.fan"),
            str(RESOURCES_ROOT / "abcd.txt"),
            "--validate",
        ]
        out, err, code = run_command(command)
        self.assertEqual(1, code)


class TestBitParsing(TestCLIParsing):
    def _test(self, example, tree, parsers, start_symbol="<start>"):
        for parser in parsers:
            parsed = False
            for actual_tree in parser.parse_multiple(example, start_symbol):
                if tree is None:
                    self.fail("Expected None")
                self.assertEqual(tree, actual_tree)
                parsed = True
                break
            if tree is None:
                self.assertTrue(True)
                return
            self.assertTrue(parsed)

    def test_bits_a(self):
        command = [
            "fandango",
            "parse",
            "-f",
            str(DOCS_ROOT / "bits.fan"),
            str(RESOURCES_ROOT / "a.txt"),
            "--validate",
        ]
        out, err, code = run_command(command)
        self.assertEqual("", err)
        self.assertEqual("", out)
        self.assertEqual(0, code)

    def test_alternative_bits(self):
        with open(RESOURCES_ROOT / "byte_alternative.fan", "r") as file:
            grammar, _ = parse(file, use_stdlib=False, use_cache=False)
            assert grammar is not None
        parser = Parser(grammar.rules)
        iter_parser = IterParsingTester(grammar.rules)
        self._test(b"\x00", None, [parser, iter_parser])
        self._test(
            b"\x01",
            DerivationTree(
                NonTerminal("<start>"),
                [
                    DerivationTree(NonTerminal("<bit>"), [DerivationTree(Terminal(0))]),
                    DerivationTree(NonTerminal("<bit>"), [DerivationTree(Terminal(0))]),
                    DerivationTree(NonTerminal("<bit>"), [DerivationTree(Terminal(0))]),
                    DerivationTree(NonTerminal("<bit>"), [DerivationTree(Terminal(0))]),
                    DerivationTree(NonTerminal("<bit>"), [DerivationTree(Terminal(0))]),
                    DerivationTree(NonTerminal("<bit>"), [DerivationTree(Terminal(0))]),
                    DerivationTree(NonTerminal("<bit>"), [DerivationTree(Terminal(0))]),
                    DerivationTree(Terminal(1)),
                ],
            ),
            [parser, iter_parser],
        )
        self._test(
            b"\x02",
            DerivationTree(
                NonTerminal("<start>"),
                [
                    DerivationTree(NonTerminal("<bit>"), [DerivationTree(Terminal(0))]),
                    DerivationTree(NonTerminal("<bit>"), [DerivationTree(Terminal(0))]),
                    DerivationTree(NonTerminal("<bit>"), [DerivationTree(Terminal(0))]),
                    DerivationTree(NonTerminal("<bit>"), [DerivationTree(Terminal(0))]),
                    DerivationTree(NonTerminal("<bit>"), [DerivationTree(Terminal(0))]),
                    DerivationTree(NonTerminal("<bit>"), [DerivationTree(Terminal(0))]),
                    DerivationTree(Terminal(1)),
                    DerivationTree(NonTerminal("<bit>"), [DerivationTree(Terminal(0))]),
                ],
            ),
            [parser, iter_parser],
        )

    def test_single_bit(self):
        with open(RESOURCES_ROOT / "bit_special.fan", "r") as file:
            grammar, _ = parse(file, use_stdlib=False, use_cache=False)
            assert grammar is not None
        parser = Parser(grammar.rules)
        iter_parser = IterParsingTester(grammar.rules)
        bit_tree_0 = DerivationTree(
            NonTerminal("<bit>"),
            [DerivationTree(Terminal(0))],
        )
        bit_tree_1 = DerivationTree(
            NonTerminal("<bit>"),
            [DerivationTree(Terminal(1))],
        )
        bit_tree_10 = DerivationTree(
            NonTerminal("<start>"),
            [
                DerivationTree(NonTerminal("<bit>"), [DerivationTree(Terminal(1))]),
                DerivationTree(NonTerminal("<bit>"), [DerivationTree(Terminal(0))]),
            ],
        )
        self._test(bit_tree_0, bit_tree_0, [parser, iter_parser], "<bit>")
        self._test(bit_tree_1, bit_tree_1, [parser, iter_parser], "<bit>")
        self._test(bit_tree_10, bit_tree_10, [parser, iter_parser], "<start>")


class TestGIFParsing(TestCLIParsing):
    def test_gif(self):
        command = [
            "fandango",
            "parse",
            "-f",
            str(DOCS_ROOT / "gif89a.fan"),
            str(DOCS_ROOT / "tinytrans.gif"),
            "--validate",
        ]
        out, err, code = run_command(command)
        self.assertEqual("", err)
        self.assertEqual("", out)
        self.assertEqual(0, code)


class TestBitstreamParsing(TestCLIParsing):
    def test_bitstream(self):
        command = [
            "fandango",
            "parse",
            "-f",
            str(RESOURCES_ROOT / "bitstream.fan"),
            str(RESOURCES_ROOT / "abcd.txt"),
            "--validate",
        ]
        out, err, code = run_command(command)
        # Warns that the number of bits (1..5) may not be a multiple of eight, # which is correct
        # self.assertEqual("", err)
        self.assertEqual("", out)
        self.assertEqual(0, code)

    def test_bitstream_a(self):
        command = [
            "fandango",
            "parse",
            "-f",
            str(RESOURCES_ROOT / "bitstream-a.fan"),
            str(RESOURCES_ROOT / "a.txt"),
            "--validate",
        ]
        out, err, code = run_command(command)
        self.assertEqual("", err)
        self.assertEqual("", out)
        self.assertEqual(0, code)

    def test_bitstream_b(self):
        command = [
            "fandango",
            "parse",
            "-f",
            str(RESOURCES_ROOT / "bitstream-a.fan"),
            str(RESOURCES_ROOT / "b.txt"),
            "--validate",
        ]
        out, err, code = run_command(command)
        # This should fail
        self.assertNotEqual("", err)
        self.assertEqual("", out)
        self.assertEqual(1, code)

    def test_rgb(self):
        command = [
            "fandango",
            "parse",
            "-f",
            str(RESOURCES_ROOT / "rgb.fan"),
            str(RESOURCES_ROOT / "rgb.txt"),
            "--validate",
        ]
        out, err, code = run_command(command)
        self.assertEqual(0, code, f"Command failed with code {code}: {err}")
        self.assertEqual("", out)
        self.assertEqual("", err)


class TestImportParsing(TestCLIParsing):
    def test_local_import(self):
        command = [
            "fandango",
            "fuzz",
            "-f",
            str(RESOURCES_ROOT / "import.fan"),
            "-n",
            "1",
        ]
        out, err, code = run_command(command)
        self.assertEqual(0, code)
        self.assertEqual("import\n", out)


class TestISO8601Parsing(TestCLIParsing):
    def test_parse_iso8601(self):
        command = [
            "fandango",
            "parse",
            "-f",
            str(DOCS_ROOT / "iso8601.fan"),
            str(RESOURCES_ROOT / "iso8601.txt"),
        ]
        out, err, code = run_command(command)
        self.assertEqual(0, code, err)
        self.assertEqual("", err)



import subprocess
import sys

import pytest


def test_run_evaluation_success():
    if sys.platform.startswith("win") or sys.platform.startswith("linux"):
        pytest.skip("bsdtar not supported on Windows and Ubuntu")

    """Test that running `python -m evaluation.run_evaluation 1` exits with code 0."""
    result = subprocess.run(
        [sys.executable, "-m", "evaluation.run_evaluation", "1"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    print("STDOUT:\n", result.stdout.decode())
    print("STDERR:\n", result.stderr.decode())

    assert result.returncode == 0, "Script did not exit cleanly"



import logging
import os
import sys
import time
import traceback
from typing import Optional, TYPE_CHECKING

from ansi_styles import ansiStyles as styles

if TYPE_CHECKING:
    from fandango.language.tree import DerivationTree
else:
    DerivationTree = object  # needs to be defined for beartype


LOGGER = logging.getLogger("fandango")
logging.basicConfig(
    level=logging.WARNING,
    format="%(name)s:%(levelname)s: %(message)s",
)

RAISE_ALL_EXCEPTIONS = os.environ.get("FANDANGO_RAISE_ALL_EXCEPTIONS")


def print_exception(e: Exception, exception_note: str | None = None):
    if RAISE_ALL_EXCEPTIONS:
        raise e

    if exception_note is not None and getattr(Exception, "add_note", None):
        # Python 3.11+ has add_note() method
        e.add_note(exception_note)
        exception_note = None

    if LOGGER.isEnabledFor(logging.INFO):
        LOGGER.info(traceback.format_exc().rstrip())
    else:
        print(type(e).__name__ + ":", e, file=sys.stderr)
        for note in getattr(e, "__notes__", []):
            print("  " + note, file=sys.stderr)

    if exception_note:
        print("  " + exception_note, file=sys.stderr)

    if "DerivationTree" in str(e):
        print(
            "  Convert <symbol> to the expected type, say 'str(<symbol>)', 'int(<symbol>)', or 'bytes(<symbol>)'",
            file=sys.stderr,
        )


USE_VISUALIZATION: bool | None = None


def set_visualization(use_visualization: bool | None):
    """Set whether to use visualization while Fandango is running"""
    global USE_VISUALIZATION
    USE_VISUALIZATION = use_visualization


COLUMNS = None
LINES = None


def use_visualization():
    """Return True if we should use visualization while Fandango is running"""
    global COLUMNS, LINES, USE_VISUALIZATION

    if COLUMNS is not None and COLUMNS < 0:
        return False  # We have checked this before

    if USE_VISUALIZATION is not None and not USE_VISUALIZATION:
        return False

    if not USE_VISUALIZATION and LOGGER.isEnabledFor(logging.INFO):
        # Don't want to interfere with logging
        COLUMNS = -1
        return False

    if not USE_VISUALIZATION and "JPY_PARENT_PID" in os.environ:
        # We're within Jupyter Notebook
        COLUMNS = -1
        return False

    if not USE_VISUALIZATION and not sys.stderr.isatty():
        # Output is not a terminal
        COLUMNS = -1
        return False

    if COLUMNS is None:
        try:
            COLUMNS, LINES = os.get_terminal_size()
        except Exception:
            if USE_VISUALIZATION:
                # Assume some reasonable default width
                COLUMNS = 80
            else:
                COLUMNS = -1
                return False

    assert COLUMNS > 0
    return True


LINE_IS_CLEAR = True


def visualize_evaluation(
    generation: int,
    max_generations: Optional[int],
    evaluation: list[tuple],
):
    """Visualize current evolution while Fandango is running"""
    if not use_visualization() or max_generations is None:
        return

    fitnesses = [fitness for _ind, fitness, _failing_trees in evaluation]
    fitnesses.sort(reverse=True)
    assert COLUMNS is not None
    columns = COLUMNS
    s = f"💃 {styles.color.ansi256(styles.rgbToAnsi256(128, 0, 0))}Fandango {styles.color.close} {generation}/{max_generations} "
    columns -= len(f"   Fandango {generation}/{max_generations} ") + 1
    columns = int(columns / 3.0)
    for column in range(0, columns):
        individual = int(column / columns * len(fitnesses))
        fitness = fitnesses[individual]

        if fitness <= 0.01:
            red, green, blue = 127, 127, 127
        else:
            red = int((1.0 - fitness) * 255)
            green = int(fitness * 255)
            blue = 0

        s += styles.bgColor.ansi256(styles.rgbToAnsi256(red, green, blue))
        s += "   "
        s += styles.bgColor.close

    global LINE_IS_CLEAR
    LINE_IS_CLEAR = False

    print(f"\r{s}", end="", file=sys.stderr)


def clear_visualization(max_generations: Optional[int] = None):
    """Clear Fandango visualization"""
    if not use_visualization():  # or max_generations is None:
        return

    global LINE_IS_CLEAR
    if LINE_IS_CLEAR:
        return

    time.sleep(0.25)
    assert COLUMNS is not None
    s = " " * (COLUMNS - 1)
    print(f"\r{s}\r", end="", file=sys.stderr)

    LINE_IS_CLEAR = True


def log_message_transfer(
    sender: str,
    receiver: str | None,
    msg: "DerivationTree",
    self_is_sender: bool,
):
    info = sender
    if receiver:
        info += f" -> {receiver}"

    if msg.contains_bits():
        print_msg = str(msg.to_bytes())
    else:
        print_msg = str(msg.value())

    LOGGER.info(f"{info}: {print_msg!r}")



from typing import Optional

from fandango.language.grammar.parser.parse_state import ParseState
from fandango.language.symbols.non_terminal import NonTerminal
from fandango.language.symbols.symbol import Symbol


class Column:
    def __init__(self, states: Optional[list[ParseState]] = None):
        self.states: list[ParseState] = states or []
        self.dot_map = dict[Symbol, list[ParseState]]()
        self.unique = set(self.states)
        for state in self.states:
            self.dot_map[state.nonterminal].append(state)

    def __iter__(self):
        yield from self.states

    def __len__(self):
        return len(self.states)

    def __getitem__(self, item):
        return self.states[item]

    def remove(self, state: ParseState):
        if state not in self.unique:
            return False
        self.unique.remove(state)
        self.states.remove(state)
        symbol = state.dot
        assert symbol is not None
        default_list: list[ParseState] = []
        self.dot_map.get(symbol, default_list).remove(state)
        return None

    def replace(self, old: ParseState, new: ParseState):
        self.unique.remove(old)
        self.unique.add(new)
        i_old = self.states.index(old)
        del self.states[i_old]
        self.states.insert(i_old, new)

        old_symbol = old.dot
        if old_symbol is not None:
            self.dot_map[old_symbol].remove(old)

        new_symbol = new.dot
        if new_symbol is not None:
            dot_list = self.dot_map.get(new_symbol, [])
            dot_list.append(new)
            self.dot_map[new_symbol] = dot_list

    def __contains__(self, item):
        return item in self.unique

    def find_dot(self, nt: NonTerminal):
        return self.dot_map.get(nt, [])

    def add(self, state: ParseState):
        if state not in self.unique:
            self.states.append(state)
            self.unique.add(state)
            symbol = state.dot
            if symbol is not None:
                state_list = self.dot_map.get(symbol, [])
                state_list.append(state)
                self.dot_map[symbol] = state_list
            return True
        return False

    def update(self, states: set[ParseState]):
        for state in states:
            self.add(state)

    def __repr__(self):
        return f"Column({self.states})"



from copy import copy
import itertools
from typing import Any, Optional
from fandango.constraints import LEGACY
from fandango.constraints.constraint import Constraint
from fandango.constraints.constraint_visitor import ConstraintVisitor
from fandango.constraints.fitness import ConstraintFitness
from fandango.language.search import NonTerminalSearch
from fandango.language.symbols.non_terminal import NonTerminal
from fandango.language.tree import DerivationTree


class ForallConstraint(Constraint):
    """
    Represents a forall constraint that can be used for fitness evaluation.
    """

    def __init__(
        self,
        statement: Constraint,
        bound: NonTerminal | str,
        search: NonTerminalSearch,
        lazy: bool = False,
        *args,
        **kwargs,
    ):
        """
        Initializes the forall constraint with the given statement, bound, and search.
        :param Constraint statement: The statement to evaluate.
        :param NonTerminal bound: The bound variable.
        :param NonTerminalSearch search: The search to use.
        :param bool lazy: If True, the forall-constraint is lazily evaluated.
        :param args: Additional arguments.
        :param kwargs: Additional keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self.statement = statement
        self.bound = bound
        self.search = search
        self.lazy = lazy

    def fitness(
        self,
        tree: DerivationTree,
        scope: Optional[dict[NonTerminal, DerivationTree]] = None,
        population: Optional[list[DerivationTree]] = None,
        local_variables: Optional[dict[str, Any]] = None,
    ) -> ConstraintFitness:
        """
        Calculate the fitness of the tree based on the given forall constraint.
        :param DerivationTree tree: The tree to evaluate.
        :param Optional[dict[NonTerminal, DerivationTree]] scope: The scope of the tree.
        :param Optional[list[DerivationTree]] population: The population of trees.
        :param Optional[dict[str, Any]] local_variables: Local variables to use in the evaluation.
        :return ConstraintFitness: The fitness of the tree.
        """
        tree_hash = self.get_hash(tree, scope, population)
        # If the fitness has already been calculated, return the cached value
        if tree_hash in self.cache:
            return copy(self.cache[tree_hash])
        fitness_values = list()
        scope = scope or dict()
        local_variables = local_variables or dict()
        # Iterate over all containers found by the search
        for container in self.search.quantify(tree, scope=scope, population=population):
            # Update the scope with the bound variable
            if isinstance(self.bound, str):
                local_variables[self.bound] = container.evaluate()
            else:
                # If the bound is a NonTerminal, update the scope
                scope[self.bound] = container.evaluate()
            # Evaluate the statement
            fitness = self.statement.fitness(tree, scope, population, local_variables)
            # Add the fitness to the list
            fitness_values.append(fitness)
            # If the forall constraint is lazy and the statement is not successful, stop
            if self.lazy and not fitness.success:
                break
        # Aggregate the fitness values
        solved = sum(fitness.solved for fitness in fitness_values)
        total = sum(fitness.total for fitness in fitness_values)
        overall = all(fitness.success for fitness in fitness_values)
        failing_trees = list(
            itertools.chain.from_iterable(
                fitness.failing_trees for fitness in fitness_values
            )
        )
        if overall:
            solved = total + 1
        total += 1
        # Create the fitness object
        fitness = ConstraintFitness(solved, total, overall, failing_trees=failing_trees)
        # Cache the fitness
        self.cache[tree_hash] = fitness
        return fitness

    def format_as_spec(self) -> str:
        bound = (
            self.bound if isinstance(self.bound, str) else self.bound.format_as_spec()
        )
        if LEGACY:
            return f"(forall {bound} in {self.search.format_as_spec()}: {self.statement.format_as_spec()})"
        else:
            return f"all({self.statement.format_as_spec()} for {bound} in {self.search.format_as_spec()})"

    def accept(self, visitor: "ConstraintVisitor"):
        """
        Accepts a visitor to traverse the constraint structure.
        :param ConstraintVisitor visitor: The visitor to accept.
        """
        visitor.visit_forall_constraint(self)
        if visitor.do_continue(self):
            self.statement.accept(visitor)



import abc
import copy
import enum
from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING, Any
from fandango.errors import FandangoValueError
from fandango.language.grammar.has_settings import HasSettings
from fandango.language.tree import DerivationTree
from fandango.logger import LOGGER

if TYPE_CHECKING:
    import fandango
    from fandango.language.grammar.node_visitors.node_visitor import NodeVisitor


class NodeType(enum.Enum):
    ALTERNATIVE = "alternative"
    CONCATENATION = "concatenation"
    REPETITION = "repetition"
    STAR = "star"
    PLUS = "plus"
    OPTION = "option"
    NON_TERMINAL = "non_terminal"
    TERMINAL = "terminal"
    CHAR_SET = "char_set"

    def __repr__(self):
        return str(self)

    def __str__(self):
        return self.value


class Node(abc.ABC):
    def __init__(
        self,
        node_type: NodeType,
        grammar_settings: Sequence["HasSettings"],
        distance_to_completion: float = float("inf"),
    ):
        self._grammar_settings = grammar_settings
        self._node_type = node_type
        self.distance_to_completion = distance_to_completion
        self._settings = NodeSettings({})
        for setting in grammar_settings:
            self._settings.update(setting.settings_for(self))

    @property
    def node_type(self):
        return self._node_type

    @property
    def is_terminal(self) -> bool:
        return self.node_type == NodeType.TERMINAL

    @property
    def is_nonterminal(self) -> bool:
        return self.node_type == NodeType.NON_TERMINAL

    @property
    def settings(self) -> "NodeSettings":
        return self._settings

    def fuzz(
        self,
        parent: DerivationTree,
        grammar: "fandango.language.grammar.grammar.Grammar",
        max_nodes: int = 100,
        in_message: bool = False,
    ):
        return

    @abc.abstractmethod
    def accept(self, visitor: "NodeVisitor"):
        raise NotImplementedError("accept method not implemented")

    def msg_parties(self, *, include_recipients: bool = False) -> set[str]:
        parties: set[str] = set()
        for child in self.children():
            parties |= child.msg_parties(include_recipients=include_recipients)
        return parties

    def slice_parties(self, parties: list[str]) -> None:
        """Remove all nodes whose party is not in `parties`."""
        for child in self.children():
            child.slice_parties(parties)

    def in_parties(self, parties: list[str]) -> bool:
        return True

    def children(self):
        return []

    def __repr__(self):
        raise NotImplementedError(
            "__repr__ not implemented, use method specific to your usecase"
        )

    def __str__(self):
        raise NotImplementedError(
            "__str__ not implemented, use method specific to your usecase"
        )

    @abc.abstractmethod
    def format_as_spec(self) -> str:
        """
        Format as a string that can be used in a spec file.
        """

    def descendents(
        self, grammar: "fandango.language.grammar.grammar.Grammar"
    ) -> Iterator["Node"]:
        """
        Returns an iterator of the descendents of this node.

        :param grammar: The rules upon which to base non-terminal lookups.
        :return An iterator over the descendent nodes.
        """
        yield from ()


NODE_SETTINGS_DEFAULTS = {
    "havoc_probability": 0.0,
    "max_stack_pow": 7,
    "terminal_should_repeat": 0.0,
    "plus_should_return_nothing": 0.0,
    "option_should_return_multiple": 0.0,
    "alternatives_should_concatenate": 0.0,
    "invert_regex": 0.0,
    "max_out_of_regex_tries": 10000,
    "non_terminal_use_other_rule": 0.0,
}


class NodeSettings:
    def __init__(
        self,
        raw_settings: dict[str, Any] = {},
    ):
        self._settings: dict[str, Any] = {}

        for k, v in NODE_SETTINGS_DEFAULTS.items():
            if k in raw_settings:
                self._settings[k] = type(v)(raw_settings[k])

    def get(self, key: str) -> Any:
        if key not in NODE_SETTINGS_DEFAULTS:
            raise FandangoValueError(f"Grammar setting {key} not found")
        if key in self._settings:
            return self._settings[key]
        else:
            return NODE_SETTINGS_DEFAULTS[key]

    def set(self, key: str, value: Any) -> None:
        if key not in NODE_SETTINGS_DEFAULTS:
            raise FandangoValueError(f"Grammar setting {key} not found")
        self._settings[key] = type(NODE_SETTINGS_DEFAULTS[key])(value)

    def __deepcopy__(self, memo: dict[int, Any]) -> "NodeSettings":
        return NodeSettings(copy.deepcopy(self._settings))

    def update(self, other: "NodeSettings | None") -> "NodeSettings":
        if other is None:
            return self

        for k in other._settings:
            if k in self._settings:
                LOGGER.warning(f"Overriding {k} with a different value")
            self._settings[k] = other._settings[k]

        return self



import abc
import enum
from typing import Any

from fandango.language.tree_value import TreeValue, TreeValueType


class SymbolType(enum.Enum):
    TERMINAL = "Terminal"
    NON_TERMINAL = "NonTerminal"
    SLICE = "Slice"
    TREE_NODE = "TreeNode"


class Symbol(abc.ABC):
    def __init__(self, value: str | bytes | int | TreeValue, type_: SymbolType):
        self._value = value if isinstance(value, TreeValue) else TreeValue(value)
        self._type = type_
        self._is_regex = False

    def check(self, word: str | int | bytes, incomplete=False) -> tuple[bool, int]:
        """Return (True, # of characters matched by `word`), or (False, 0)"""
        return False, 0

    def check_all(self, word: str | int | bytes) -> bool:
        """Return True if `word` matches"""
        return False

    @property
    def is_terminal(self) -> bool:
        return self._type == SymbolType.TERMINAL

    @property
    def is_non_terminal(self) -> bool:
        return self._type == SymbolType.NON_TERMINAL

    @property
    def is_slice(self) -> bool:
        return self._type == SymbolType.SLICE

    @property
    def is_regex(self) -> bool:
        return getattr(self, "_is_regex", False)

    def is_type(self, type_: TreeValueType) -> bool:
        return self._value.is_type(type_)

    def value(self) -> TreeValue:
        return self._value

    def __eq__(self, other: Any) -> bool:
        return type(self) is type(other) and self._value == other._value

    @abc.abstractmethod
    def __hash__(self) -> int:
        return NotImplemented

    def _repr(self) -> str:
        return str(self.value())

    def __str__(self) -> str:
        raise KeyError(f"str() not implemented for {type(self)}, use specific function")

    def __repr__(self) -> str:
        raise KeyError(
            f"repr() not implemented for {type(self)}, use specific function"
        )
